
//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	pointingfilter.cc --- Stable pointing data filtering
//
//
//
// CHANGE HISTORY:
//   AI, 2012-09-18: 1.0.0: - inital version
//
//
//	AUTHOR:
//		A. Ibarra, Tue Sep  18 15:15:45 MEST 2012
//

#include "pointingfilter.h"
#include "epframes.h"
#include "emframes.h"
#include "rgsframes.h"
#include "Tracer.h"
#include "errstr.h"
#include "param.h"
#include "OalServer.h"
#include "Proc.h"
#include "DataFilter.h"
#include "TimeOrderedTable.h"

//gcc4.8.2 portability changes
#include <cstdio>

class FrameIdentifier {
public:

	FrameIdentifier(){};

	void runepframes(const OdfSet* os)
	{

		errstr << "Running epframes... " << Sas::message(Msg::App, Msg::Sparse);
		Proc p("./");

		Info::Task* epframes = task("epframes");
		ParameterInfo& epframespars = *epframes->parameters();
		epframespars.revertToDefault();
		std::stringstream name;
		name << os->dirname()<<"/"<<os->basename();
		epframespars["set"]->setValue(name.str());
		epframespars["gtiset"]->setValue("gtitmp.ds");

		_tmpName << tmpnam(0) << Xmm::toString(os->instrument()) << os->exposureNr()
				<< os->ccdNr()<< ".ds"  ;

		epframespars["eventset"]->setValue(_tmpName.str());


		if(!p.protectedrun(epframes))
			throw(ProcException(""));

	}


	void runemframes(const OdfSet* os)
	{
		errstr << "Running emframes... " << Sas::message(Msg::App, Msg::Sparse);
		Proc p("./");

		oalServer << os->name();


		Info::Task* emframes = task("emframes");
		ParameterInfo& emframespars = *emframes->parameters();
		emframespars.revertToDefault();
		if(oalServer.odf()->hasSet(Xmm::AUXILIARY))
			emframespars["auxiliaryset"]->setValue(oalServer.odf()->selectSet(Xmm::AUXILIARY)->name());
		else
			throw(ProcException(""));

		_tmpName <<tmpnam(0) << Xmm::toString(os->instrument()) << os->exposureNr()
				<< os->ccdNr()<< ".ds"  ;

		emframespars["frameset"]->setValue(_tmpName.str());
		emframespars["withodfeventset"]->setValue(true);
		std::stringstream scienceFileName;
		scienceFileName << os->dirname()<<"/"<<os->basename();
		emframespars["odfeventset"]->setValue(scienceFileName.str());
		emframespars["neweventset"]->setValue(true);
		_KKK = new stringstream();

		*_KKK << tmpnam(0) << Xmm::toString(os->instrument()) << os->exposureNr() <<"_Evt_CCD_"
						<< os->ccdNr()<< ".ds"  ;
		emframespars["outeventset"]->setValue(_KKK->str());
		emframespars["writegtiset"]->setValue(false);
		emframespars["flagbadtimes"]->setValue(false);



		if(!p.protectedrun(emframes))
			throw(ProcException(""));

	}


	void runrgsframes(const OdfSet* os)
	{
		errstr << "Running rgsframes... " << Sas::message(Msg::App, Msg::Sparse);
		Proc p("./");

		oalServer << os->name();

		_tmpName << tmpnam(0) <<Xmm::toString(os->instrument()) << os->exposureNr()
				<< os->ccdNr()<< ".ds"  ;


		Info::Task* rgsframes = task("rgsframes");
		ParameterInfo& rgsframespars = *rgsframes->parameters();
		rgsframespars.revertToDefault();

		rgsframespars["ccdset"]->setValue(_tmpName.str());

		rgsframespars["set"]->setValue(os->name());
		switch ( os->instrument() ) {
			case Xmm::RGS1: rgsframespars["instrument"]->setValue("rgs1"); break;
			case Xmm::RGS2: rgsframespars["instrument"]->setValue("rgs2"); break;
			default:
				errstr << "unexpected instrument " << Xmm::toString(os->instrument())
				<< " for Machine::exposureProc"
				<< Sas::error("InternalError");
		}

		if(!p.protectedrun(rgsframes))
			throw(ProcException(""));
	}

	void identifyFrames(const OdfSet* os,real64 start, real64 stop)
	{
		errstr << "Identifying frames in: " << _tmpName.str() << Sas::message(Msg::App, Msg::Sparse);

		std::stringstream tab;
		if (os->instrument() == Xmm::EPN)
			tab << "EVENTS";
		else if ( (os->instrument() == Xmm::EMOS1) || (os->instrument() == Xmm::EMOS2) )
			tab << "FRAMES";
		else
			tab << "EXPOSURE";

		_frameTable = dataSetServer->open(_tmpName.str(),Dal::Modify)->table(tab.str());


		// if MOS substract the telemetry drops..
		int badFramesBeforeCurrentStablePointing = 0;
		int badFramesCurrentStablePointing = 0;
		if ( (os->instrument() == Xmm::EMOS1) || (os->instrument() == Xmm::EMOS2) )
		{
			errstr << "START TIME " << oalServer.odf()->currentExposure()->actualStartTime().toString() << Sas::message(Msg::App, Msg::Sparse);
			double ts = oalServer.odf()->currentExposure()->actualStartTime() - Xmm::missionReferenceTime ;
			errstr << "START TIME " << setprecision(14) << ts	<< Sas::message(Msg::App, Msg::Sparse);

			if (ts > start )
			{
				errstr << "Stable pointing starts before current exposure. Set the stable pointing start " <<
						"time to the current exposure start time "  << Sas::message(Msg::App, Msg::Sparse);
				start = ts;

			}
			std::stringstream strBeforeFlag;
			strBeforeFlag<< std::setprecision(19) << "(FLAG != 0) && TIME in [ "<<ts << ':' << start << ']';
			errstr << "Previous filter expression " << strBeforeFlag.str() << Sas::message(Msg::App, Msg::Sparse);
			DataFilter flagBeforeFilter(strBeforeFlag.str());
			badFramesBeforeCurrentStablePointing = flagBeforeFilter.doSelect(*_frameTable);

			std::stringstream strCurrentFlag;
			strCurrentFlag<< std::setprecision(19) << "(FLAG != 0) && TIME in [ "<<start << ':' << stop << ']';
			errstr << "Current filter expression " << strCurrentFlag.str() << Sas::message(Msg::App, Msg::Sparse);
			DataFilter flagCurrentFilter(strCurrentFlag.str());
			badFramesCurrentStablePointing = flagCurrentFilter.doSelect(*_frameTable);

			errstr << "MOS frames lost (telemetry drops, gaps...) before current stable pointing " <<
					badFramesBeforeCurrentStablePointing << Sas::message(Msg::App, Msg::Sparse);

			errstr << "MOS frames lost (telemetry drops, gaps...) during current stable pointing " <<
					badFramesCurrentStablePointing << Sas::message(Msg::App, Msg::Sparse);

		}


		std::stringstream str;
		str<< std::setprecision(19) << "TIME in [ "<<start << ':' << stop << ']';

		DataFilter gtiFilter(str.str());

		gtiFilter.doFilter(*_frameTable);

		if ( _frameTable->rows() > 0)
		{
			ColumnData *frameCol = _frameTable->column("FRAME")->data();
			int32 *frameData = 0;
			frameData = frameCol->int32Data();
			_frameStart = frameData[0];
			_frameStop = frameData[_frameTable->rows()-1 ];
			setEMOSStartFrame(frameData[0]);
			setEMOSStopFrame(frameData[_frameTable->rows()-1 ]);
			if ( (os->instrument() == Xmm::EMOS1) || (os->instrument() == Xmm::EMOS2) )
			{
//				cout << "FRAME WHERE I SHOULD START " <<  _frameStart << "  STOP " <<  _frameStop << endl;
//				cout << "Open Evt file and identify the row for this frames " << endl;
				string tabName;
				if (os->instrument() == Xmm::EMOS1)
					tabName = "M1IME1";
				else
					tabName = "M2IME1";

				Table *evtFrameTab = dataSetServer->open(_KKK->str(),Dal::Read)->table(tabName);
				ColumnData *evtFrameCol = evtFrameTab->column("FRAME")->data();
				int32 *evtFrameData = 0;
				evtFrameData = evtFrameCol->int32Data();
				unsigned int starti= 0;

				while ((evtFrameData[starti] != _frameStart) && (starti <= evtFrameTab->rows()-1))
					starti++;

				//It could happend that this frame does not exist in the event file. Therefore, we have to find the frame
				//close the input frame

				if (starti == evtFrameTab->rows()&& (_frameStart != evtFrameData[starti-1]))
				{
					starti=0;
					while ((evtFrameData[starti] != _frameStart) && (starti <= evtFrameTab->rows()-1))
					{
						//cout << "SIGN " << _frameStart << " " <<evtFrameData[starti] << " RES "  << _frameStart - evtFrameData[starti] << endl;
						if ( (_frameStart - evtFrameData[starti]) < 0 )
						{
							starti--;
							errstr << "closed frame found " << evtFrameData[starti] << Sas::message(Msg::App, Msg::Sparse);

							break;
						}
						starti++;
					}
				}

				if (starti == evtFrameTab->rows())
					errstr << "Frame matching error" << Sas::error("NoFrameFound");


				unsigned int stopi = 0;
				while (  (evtFrameData[stopi] != _frameStop) && (stopi <= evtFrameTab->rows()-1))
					stopi++;


				//It could happend that this frame does not exist in the event file. Therefore, we have to find the frame
				//close the input frame

				if ( (stopi == evtFrameTab->rows() ) && (_frameStop != evtFrameData[stopi-1]) )
				{
					stopi=0;
					while ((evtFrameData[stopi] != _frameStop) && (stopi <= evtFrameTab->rows()-1))
					{
						//cout << "SIGN " << _frameStop << " " <<evtFrameData[stopi] << " RES "  << _frameStop - evtFrameData[stopi] << endl;
						if ( (_frameStop - evtFrameData[stopi]) < 0 )
						{
							stopi--;
							errstr << "closed frame found " << evtFrameData[stopi] << Sas::message(Msg::App, Msg::Sparse);

							break;
						}
						stopi++;
					}
					// We reached the end of the file and we didn't find the frame number...
					if (stopi == evtFrameTab->rows())
					{
						errstr << "End of the file reached and no frame found in event file.  "<< Sas::message(Msg::App, Msg::Sparse);
						errstr << "Assigning the last frame value... "<< Sas::message(Msg::App, Msg::Sparse);
						_frameStop = evtFrameData[stopi-1];
						stopi = evtFrameTab->rows()-2;
					}

				}

				if (stopi == evtFrameTab->rows())
					errstr << "Frame matching error" << Sas::error("NoFrameFound");



				std::stringstream scienceFileName;
				scienceFileName << os->dirname()<<"/"<<os->basename();
				delete evtFrameCol;
//				cout << "EMOS RAW IME file name " << scienceFileName.str() << endl;
				ColumnData *rawEvtFrameCol = dataSetServer->open(scienceFileName.str(),Dal::Read)->table(tabName)->column("FRAME")->data();
				int32 *rawEvtFrameData = 0;
				rawEvtFrameData = rawEvtFrameCol->int32Data();

//				cout << "ORIGINAL START FRAME " << rawEvtFrameData[starti+1] << endl;
//				cout << "ORIGINAL STOP FRAME " << rawEvtFrameData[stopi+1] << endl;
				_frameStart = rawEvtFrameData[starti+1] ;
				_frameStop = rawEvtFrameData[stopi+1] ;
				if (_frameStart < getEMOSStartFrame())
				{
					errstr << "First frame in the evt file "<< _frameStart << " is smaller than the first frame in the AUX "
							<< getEMOSStartFrame() << " Re-setting the first frame of AUX file " << Sas::message(Msg::App, Msg::Sparse);
					setEMOSStartFrame(_frameStart);
				}
//				cout << "FRAMES FOR AUX " << getEMOSStartFrame() << " " << getEMOSStopFrame() << endl;
				delete rawEvtFrameCol;
				//_frameStart = _frameStart - badFramesBeforeCurrentStablePointing;
				//_frameStop = _frameStop -
				//		(badFramesBeforeCurrentStablePointing + badFramesCurrentStablePointing);
				dataSetServer -> remove (_KKK->str());
				delete _KKK;
			}

			errstr << "Taking into account the frame lost: FIRST FRAME: " << _frameStart
					<< " LAST  FRAME: " << _frameStop << Sas::message(Msg::App, Msg::Sparse);

			delete frameCol;
		}
		dataSetServer -> remove(_tmpName.str());
		_tmpName.str("");

	}


	bool checkExposure(const OdfSet* os,real64 start, real64 stop)
	{


	}



	void filterRawfile(const OdfSet* os, int pointing) {

		std::stringstream rawFileName;
		rawFileName << os->dirname()<<"/"<<os->basename();
		errstr << "Filtering raw file: " << rawFileName.str()  << Sas::message(Msg::App, Msg::Sparse);

		//Pointing number
		std::stringstream pointingstr;
		pointingstr << setw(2) << setfill('0') << pointing;

		std::stringstream pointingRawFileName;
		pointingRawFileName << os->basename();
		std::string rawFile = pointingRawFileName.str().replace(13,2,pointingstr.str());
		DataSet *ds = dataSetServer->clone(rawFileName.str(),
				rawFile,Dal::Modify);
		std::stringstream inst;
		if (os->instrument() == Xmm::EPN)
			inst << "PNIME1";
		else if (os->instrument() == Xmm::EMOS1)
			inst << "M1IME1";
		else if (os->instrument() == Xmm::EMOS2)
			inst << "M2IME1";
		else if (os->instrument() == Xmm::RGS1)
			inst << "R1SPE1";
		else if (os->instrument() == Xmm::RGS2)
			inst << "R2SPE1";

		Table *rawTable = ds->table(inst.str());
		std::stringstream str;
		str<< std::setprecision(19) << "FRAME in [ "<<_frameStart << ':' << _frameStop << ']';

		DataFilter frameFilter(str.str());

//		STime t = Xmm::missionReferenceTime + start;

//		*(rawTable)->attribute("DATE-OBS") =  t.toString(0, STime::FITS);
//		t = Xmm::missionReferenceTime + stop;
//		*(rawTable)->attribute("DATE-END") = t.toString(0, STime::FITS);

		frameFilter.doFilter(*rawTable);

		ColumnData *frameCol = rawTable->column("FRAME")->data();
		int32 *frameData = 0;
		frameData = frameCol->int32Data();
//		cout << "FINAL FRAME CUT:  START --> " <<  frameData[0] << "  STOP --> " << frameData[rawTable->rows()-1 ] << endl;
		delete frameCol;
		dataSetServer->close(ds);
	}

	std::string createAUXexpression(int ccdNr,Xmm::Instrument inst )
	{

		std::stringstream str;

		if ( inst ==  Xmm::EPN )
		{
			ccdNr=ccdNr-1;
			int quadrant;
			switch(ccdNr){
				case 0:
					ccdNr = 0;
					quadrant = 0;
					break;
				case 1:
					ccdNr = 1;
					quadrant = 0;
					break;
				case 2:
					ccdNr = 2;
					quadrant = 0;
					break;
				case 3:
					ccdNr = 0;
					quadrant = 1;
					break;
				case 4:
					ccdNr = 1;
					quadrant = 1;
					break;
				case 5:
					ccdNr = 2;
					quadrant = 1;
					break;
				case 6:
					ccdNr = 0;
					quadrant = 2;
					break;
				case 7:
					ccdNr = 1;
					quadrant = 2;
					break;
				case 8:
					ccdNr = 2;
					quadrant = 2;
					break;
				case 9:
					ccdNr = 0;
					quadrant = 3;
					break;
				case 10:
					ccdNr = 1;
					quadrant = 3;
					break;
				case 11:
					ccdNr = 2;
					quadrant = 3;
					break;
			default:
					errstr << "InvalidCCDNR" << Sas::error("InvalidCCDNR");
			}
			str<< std::setprecision(19) << " ( (FRAME in [ "<<_frameStart << ':' << _frameStop << "]) && (QUADRANT == " <<
				quadrant << ") && (CCDID == " << ccdNr << ") ) ";
		}
		else if ( (inst ==  Xmm::EMOS1) || (inst ==  Xmm::EMOS2) )
		{
			_frameStart = getEMOSStartFrame();
			_frameStop = getEMOSStopFrame();
			str<< std::setprecision(19) << " ( (FRAME in [ "<<_frameStart << ':' << _frameStop << "]) && (CCDID == " << ccdNr << ") ) ";
		}
		else
		{
			str<< std::setprecision(19) << " ( (FRAME in [ "<<_frameStart << ':' << _frameStop << "]) && (CCDID == " << ccdNr << ") ) ";
		}

		return str.str();
	}


	void filterAUXfile(const OdfSet *odfSet,int pointing, string expr)
	{
		std::stringstream auxFileName;
		auxFileName << odfSet->dirname()<<"/"<<odfSet->basename();
		errstr << "Filtering AUX file: " << auxFileName.str()  << Sas::message(Msg::App, Msg::Sparse);
		errstr << "Filter expression... " << expr << Sas::message(Msg::App, Msg::Sparse);
		std::stringstream pointingstr;
		pointingstr << std::setw(2) << std::setfill('0') << pointing;

		std::stringstream pointingAUXFileName;
		pointingAUXFileName << odfSet->basename();
		std::string auxFile = pointingAUXFileName.str().replace(13,2,pointingstr.str());
		//cout << "NEW AUX NAMe " << auxFile << endl;
		DataSet *ds;
		if (!dataSetServer->exists(auxFile))
			ds = dataSetServer->clone(auxFileName.str(),auxFile,Dal::Modify);
		else
		     ds =  dataSetServer->open(auxFile.c_str(),Dal::Modify);

		std::stringstream inst;
		if (odfSet->instrument() == Xmm::EPN)
			inst << "PNAUX1";
		else if (odfSet->instrument() == Xmm::EMOS1)
			inst << "M1AUX1";
		else if (odfSet->instrument() == Xmm::EMOS2)
			inst << "M2AUX1";
		else if (odfSet->instrument() == Xmm::RGS1)
			inst << "R1AUX1";
		else if (odfSet->instrument() == Xmm::RGS2)
			inst << "R2AUX1";

		Table *rawTable = ds->table(inst.str());
		DataFilter gtiFilter(expr);
		unsigned long selectedEvents =  gtiFilter.doFilter(*rawTable);
		errstr << "number of AUX frames filtered " << selectedEvents << Sas::message(Msg::App, Msg::Sparse);
		dataSetServer->close(ds);
	}


	void filterHKfile(const OdfSet *odfSet,int pointing, STime start, STime stop)
	{
		std::stringstream auxFileName;
		auxFileName << odfSet->dirname()<<"/"<<odfSet->basename();
		errstr << "Filtering HK file: " << auxFileName.str()  << Sas::message(Msg::App, Msg::Sparse);

		std::stringstream pointingstr;
		pointingstr << setw(2) << setfill('0') << pointing;

		std::stringstream pointingAUXFileName;
		pointingAUXFileName << odfSet->basename();
		std::string auxFile = pointingAUXFileName.str().replace(13,2,pointingstr.str());

		DataSet *ds;
		if (!dataSetServer->exists(auxFile))
			ds = dataSetServer->clone(auxFileName.str(),auxFile,Dal::Modify);
		else
		     ds =  dataSetServer->open(auxFile.c_str(),Dal::Modify);

		std::stringstream inst;
		if (odfSet->instrument() == Xmm::EPN)
			inst << "PNPMH1";
		else if (odfSet->instrument() == Xmm::EMOS1)
			inst << "M1PEH1";
		else if (odfSet->instrument() == Xmm::EMOS2)
			inst << "M2PEH1";
		else if (odfSet->instrument() == Xmm::RGS1)
			inst << "R1PFH1";
		else if (odfSet->instrument() == Xmm::RGS2)
			inst << "R2PFH1";


		Table* tb = ds->table(inst.str());

		TimeOrderedTable * tot = new TimeOrderedTable(tb);
		unsigned int firstRow = tot->lower_bound(start) ;
		unsigned int lastRow = tot->upper_bound(stop) ;
		//cout << "ROWS1 " << tot->rows() << endl;
		//cout << "START "<< start << " " << firstRow << " "<< tot->time_of_row(tot->lower_bound(start)) << endl;
		//cout << "ROWS2 " << tot->rows() << endl;
		//cout << "STOP " << stop << " " << lastRow << " "<< tot->time_of_row(tot->upper_bound(stop)) << endl;

		unsigned rows = tot->rows()-1;
		delete tot;
		tb->deleteRows(0,firstRow);
		tb->deleteRows(lastRow-firstRow+1,rows);


		dataSetServer->close(ds);

	}

	void setEMOSStartFrame(int32 startFrame)
	{
		_emosStartFrame = startFrame;
	}

	int32 getEMOSStartFrame()
	{

		return _emosStartFrame;

	}

	void setEMOSStopFrame(int32 stopFrame)
	{
		_emosStopFrame = stopFrame;
	}

	int32 getEMOSStopFrame()
	{
		return _emosStopFrame;

	}


private:
	Table *_frameTable;
	std::stringstream _tmpName;
	std::stringstream *_KKK;
	int32 _frameStart;
	int32 _frameStop;
	int32 _emosStartFrame;
	int32 _emosStopFrame;

};



void pointingfilter()
{


	Table *gtiTable = dataSetServer->open(stringParameter("preqgtifile"),Dal::Read)
			->table("STDGTI");


	ColumnData *startCol = gtiTable->column("START")->data();
	ColumnData *stopCol = gtiTable->column("STOP")->data();
	real64 *startData=0;
	real64 *stopData=0;
	startData = startCol->real64Data();
	stopData = stopCol->real64Data();

	FrameIdentifier fi;

	std::vector<Xmm::Instrument > instruments;
	if (booleanParameter("withPN"))
		instruments.push_back(Xmm::EPN);
	if(booleanParameter("withMOS1"))
		instruments.push_back(Xmm::EMOS1);
	if(booleanParameter("withMOS2"))
	instruments.push_back(Xmm::EMOS2);
	if(booleanParameter("withRGS1"))
		instruments.push_back(Xmm::RGS1);
	if(booleanParameter("withRGS2"))
		instruments.push_back(Xmm::RGS2);


//	std::vector<Xmm::DataMode> dataModes ;
//	dataModes.push_back(Xmm::IMAGING);
//	dataModes.push_back(Xmm::BURST);
//	dataModes.push_back(Xmm::TIMING);
//	dataModes.push_back(Xmm::COMPRESSED_TIMING);
//	dataModes.push_back(Xmm::REDUCED_IMAGING);


	//Initial pointing value!!!! By definition!!!
	unsigned int pointingCounter = 31;
	unsigned long rows = gtiTable->rows();

	unsigned int pointing = 0;
	if (booleanParameter("withpointingid"))
	{
		pointing = intParameter("pointingid") -1 ;
		rows = pointing+1;
		pointingCounter = pointingCounter + pointing;
	}

	for (unsigned int p = pointing ; p < rows ; ++p)
	{
		if ( pointing >= gtiTable->rows())
		{
			errstr << "Requested stable pointing does not exists" <<  Sas::error("UnknownStablePointing");

		}
		errstr << "Processing Pointing number: " << pointingCounter << Sas::message(Msg::App, Msg::Sparse);

		for(unsigned l = 0; l <= instruments.size()-1; ++l){
			oalServer.instrument(instruments[l]);
			//cout << Xmm::toString(instruments[l]) <<endl;
			if ((instruments[l] == Xmm::RGS1) || (instruments[l] == Xmm::RGS2) )
				oalServer.dataMode(Xmm::DATA_SPECTROSCOPY);
			else
				oalServer.dataMode(Xmm::IMAGING);

			unsigned N = oalServer.odf()->exposures();
			errstr << "Considering one of the " << N << " exposures for mode "
					<< oalServer.dataMode() << Sas::message(Msg::Meta, Msg::Silent);

			errstr << "Number of exposures: " << N  << Sas::message(Msg::App, Msg::Sparse);



			for(unsigned n = 0; n < N; n++){
				oalServer.exposureIndex(n+1);

				std::string id = oalServer.instrExposureId();

				errstr << "Exposure " << id   << Sas::message(Msg::App, Msg::Sparse);
				errstr << "Considering exposure " << id
						<< "[index=" << n << "]" << Sas::message(Msg::Meta, Msg::Silent);

				const list<const OdfSet*>& l = oalServer.odf()->currentExposure()->sets();
				typedef std::list<const OdfSet*>::const_iterator LI;

				//Check if the stable pointing affects the current exposure
				errstr << "Stable pointing interval: [ " << startData[p]+Xmm::missionReferenceTime << " : "
						<<  stopData[p]+Xmm::missionReferenceTime << " ] " << Sas::message(Msg::App, Msg::Sparse);
				errstr << "Exposure interval:        [ " << oalServer.odf()->currentExposure()->actualStartTime().toString() << " : "
						<< oalServer.odf()->currentExposure()->actualEndTime().toString() << " ] " << Sas::message(Msg::App, Msg::Sparse);

				if ( (startData[p]+Xmm::missionReferenceTime) >  oalServer.odf()->currentExposure()->actualEndTime() &&
						(startData[p]+Xmm::missionReferenceTime) > oalServer.odf()->currentExposure()->actualStartTime() )

				{
					errstr << "skipping expo: Exposure ends before stable pointing starts" << Sas::message(Msg::App, Msg::Sparse);
					//break;
				}
				else if ( (oalServer.odf()->currentExposure()->actualStartTime() > stopData[p]+Xmm::missionReferenceTime ) &&
						( oalServer.odf()->currentExposure()->actualEndTime() > stopData[p]+Xmm::missionReferenceTime ) )
				{
					errstr << "skipping expo: Exposure starts after stable pointing ends" << Sas::message(Msg::App, Msg::Sparse);
				}
				else
				{
					errstr << "PROCESSING... " << Sas::message(Msg::App, Msg::Sparse);
					std::stringstream expr;

					for(LI i=l.begin(); i!=l.end(); ++i){
						const OdfSet* os = (*i);

						if (os->instrument() == Xmm::EPN)
						{
							fi.runepframes(os);
						}
						else if ((os->instrument() == Xmm::EMOS1) ||
								(os->instrument() == Xmm::EMOS2) )
						{
							fi.runemframes(os);
						}
						else
							fi.runrgsframes(os);

						//cout << setprecision(19)<<"TIME " << startData[p] << " " <<
						//		stopData[p] << endl;

						fi.identifyFrames(os,startData[p],stopData[p]);

						fi.filterRawfile(os,pointingCounter);
						expr << fi.createAUXexpression(os->ccdNr(),os->instrument());
						LI j =i;
						if (++j != l.end() )
							expr << " || ";

					}

					if (oalServer.odf()->hasSet(Xmm::AUXILIARY) )
					{
						const OdfSet *odfSet = oalServer.odf()->selectSet(Xmm::AUXILIARY);
						fi.filterAUXfile(odfSet,pointingCounter,expr.str());
					}
				}

			} // N
		} //datamodes


		//Filter Housekeeping data---> Not longer needed!!!!
/*
		for(unsigned l = 0; l <= instruments.size()-1; ++l){
			oalServer.instrument(instruments[l]);
			//cout << Xmm::toString(instruments[l]) <<endl;
			const OdfSet *odfSet;
			if ((instruments[l] == Xmm::RGS1) || (instruments[l] == Xmm::RGS2) )
			{
				oalServer.dataMode(Xmm::FULL_PERIODIC_HOUSEKEEPING);
				odfSet = oalServer.odf()->selectSet(Xmm::FULL_PERIODIC_HOUSEKEEPING);
			}
			else if((instruments[l] == Xmm::EMOS1) || (instruments[l] == Xmm::EMOS2) )
			{
				oalServer.dataMode(Xmm::PERIODIC_HOUSEKEEPING);
				odfSet = oalServer.odf()->selectSet(Xmm::PERIODIC_HOUSEKEEPING);
			}
			else
			{
				oalServer.dataMode(Xmm::MAIN_PERIODIC_HOUSEKEEPING);
				odfSet = oalServer.odf()->selectSet(Xmm::MAIN_PERIODIC_HOUSEKEEPING);
			}
			cout <<"FILTERING HOUSEKEEPING2" << endl;
			unsigned N = oalServer.odf()->exposures();
			errstr << "Considering one of the " << N << " exposures for mode "
					<< oalServer.dataMode() << Sas::message(Msg::Meta, Msg::Silent);

			errstr << "Number of exposures: " << N  << Sas::message(Msg::App, Msg::Sparse);
			cout <<"FILTERING HOUSEKEEPING3" << endl;


			for(unsigned n = 0; n < N; n++){
				oalServer.exposureIndex(n+1);

				std::string id = oalServer.instrExposureId();

				errstr << "Exposure " << id   << Sas::message(Msg::App, Msg::Sparse);
				errstr << "Considering exposure " << id
						<< "[index=" << n << "]" << Sas::message(Msg::Meta, Msg::Silent);

				STime start = startData[p]+Xmm::missionReferenceTime;
				STime stop = stopData[p]+Xmm::missionReferenceTime;

				fi.filterHKfile(odfSet,pointingCounter,start, stop);
			}

		}
*/
		pointingCounter++;
	}

	  delete startCol;
	  delete stopCol;

}
