//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	preqgti.cc --- Stable pointing data filtering and GTI creation
//
//
//
// CHANGE HISTORY:
//   AI, 2012-08-27: 1.0.0: - inital version
//
//
//	AUTHOR:
//		A. Ibarra, Mon Aug  27 15:15:45 MEST 2012
//


#include "preqgti.h"
#include <cmath>

#include "stringutil.h"
#include <Tracer.h>
#include "errstr.h"
#include "param.h"
#include "OalServer.h"
#include "OdfInfo.h"
#include "atthkgen.h"
#include "AttitudeLocator.h"
#include <SasDalUtils.h>


#include <SasDalUtils.h>
#include "GtiTableFilter.h"

//gcc4.8.2 portability changes
#include <cstdio>
#include <unistd.h>

using namespace std;

class StablePointngGtiTable {
public:
	StablePointngGtiTable()
	{

		//	inquire exposure start/end times
		_obsTStart = oalServer.odf()->observationStartTime();
		_obsTStop = oalServer.odf()->observationEndTime();
		_ATSFile = oalServer.odf()->selectSet(Xmm::SC, Xmm::ATTITUDE_HISTORY)->name();
		errstr <<"Exposure window: ["<< _obsTStart <<", "
			  << _obsTStop<<']'
			  << Sas::message(Msg::App, Msg::Sparse);

	}


	StablePointngGtiTable(string atsfile)
	{

		//	inquire exposure start/end times
		//_obsTStart = oalServer.odf()->observationStartTime();
		//_obsTStop = oalServer.odf()->observationEndTime();
		_ATSFile = atsfile;
		DataSet *set = dataSetServer->open(_ATSFile,Dal::Read);
		Table *ds = set->table(0);

		if (ds->hasAttribute("DATE-OBS") )
		{
			STime st(ds->attribute("DATE-OBS")->asString());
			_obsTStart = st;
		}

		if (ds->hasAttribute("DATE-END") )
		{
			STime st(ds->attribute("DATE-END")->asString());
			_obsTStop = st;
		}

		dataSetServer->close(set);

		errstr <<"Exposure window: ["<< _obsTStart <<", "
			  << _obsTStop<<']'
			  << Sas::message(Msg::App, Msg::Sparse);

	}


	StablePointngGtiTable(real64 timebegin, real64 timeend)
	{

		//	inquire exposure start/end times
		_obsTStart = timebegin + Xmm::missionReferenceTime;
		_obsTStop = timeend + Xmm::missionReferenceTime;
		_ATSFile = oalServer.odf()->selectSet(Xmm::SC, Xmm::ATTITUDE_HISTORY)->name();
		errstr <<"Exposure window: ["<< _obsTStart <<", "
			  << _obsTStop<<']'
			  << Sas::message(Msg::App, Msg::Sparse);

	}


	StablePointngGtiTable(string atsfile, real64 timebegin, real64 timeend)
	{

		//	inquire exposure start/end times
		_obsTStart = timebegin + Xmm::missionReferenceTime;
		_obsTStop = timeend + Xmm::missionReferenceTime;
		_ATSFile = atsfile;
		errstr <<"Exposure window: ["<< _obsTStart <<", "
			  << _obsTStop<<']'
			  << Sas::message(Msg::App, Msg::Sparse);

		errstr << "Exposure window: ["<< _obsTStart <<", "
						  << _obsTStop<<']'
						  <<  Sas::message(Msg::App, Msg::Sparse);
	}


void createGtiTable()
{
	stringstream expr;
	expr << "(TYPEID != \"C\") && (VALTIME > \"" << _obsTStart <<  "\") && (VALTIME < \""<<  _obsTStop << "\")";
	//	run tabgtigen
	errstr << "tabgtigen filter expression " << expr.str() <<Sas::message(Msg::App, Msg::Sparse);

	Info::Task* tabgtigen = task("tabgtigen");
	ParameterInfo& tabgtigenpars = *tabgtigen->parameters();
	tabgtigenpars["table"]->setValue(_ATSFile);
	tabgtigenpars["gtiset"]->setValue(stringParameter("preqgtifile"));
	tabgtigenpars["expression"]->setValue(expr.str());
	tabgtigenpars["timecolumn"]->setValue("VALTIME");

	tabgtigen->exec();
}

double angularSeparation(vector<double> A, vector<double> B )
{
	return acos(sin(A[1])*sin(B[1])+cos(A[1])*cos(B[1])*cos(B[0]-A[0]));
}



void getPReqValues()
{

	DataSet *s = dataSetServer->open(_ATSFile, Dal::Read);

	string extName = "SCATS1";
	if (!s->hasBlock("SCATS1"))
		extName="SCATS49";


	//Open the original ATS file to get the correct VALDUR value

	Table *setT = s->table(extName);
	ColumnData *valdurCol = setT->column("VALDUR")->data();
	real32 *valdur = valdurCol->real32Data();

	Column *valtimeCol = setT->column("VALTIME");
	ColumnData *valtime = valtimeCol->data();


	string atsSetName = tmpnam(0);
	DataSet *atsDS = dataSetServer->clone(_ATSFile,
			atsSetName,Dal::Modify);



	Table *atsTab = atsDS->table(extName);
	Column *c = atsTab->column("PREQID");


	_obsTStart = findClosestValTime(_obsTStart, valtime, valtimeCol->rows(), valdur);
	errstr << "NEW STARTT TIME: " << _obsTStart <<  Sas::message(Msg::App, Msg::Sparse);

	std::stringstream expr;

	//expr << " (VALTIME > \"" << _obsTStart <<  "\") && (VALTIME < \""<<  _obsTStop << "\")";
	//(ahf.valtime) le Obs_end and (ahf.valtime+valdur) ge Obs_Start
	_obsTStart = _obsTStart - 1.;
	expr << " (VALTIME <= \"" << _obsTStop <<  "\") && (VALTIME >= \""<<  _obsTStart << "\")";
	//expr << " (VALTIME <= \"" << _obsTStop <<  "\") && (VALTIME >= \"2014-01-18T13:37:48.000\" )";
	DataFilter obsFilter(expr.str());

	errstr << "Filter expression:  " << expr.str() <<  Sas::message(Msg::App, Msg::Sparse);

	//unsigned long selectedEvents = obsFilter.doFilter(*atsTab);
	vector<DataFilter::RowRange> rowRangeVec;
	obsFilter.doSelect(*atsTab,rowRangeVec);

	//errstr << "selected " << selectedEvents  <<  " rows from the input table." << Sas::message(Msg::App, Msg::Sparse);
	ColumnData *cd = c->data();

	string preq = "";
	std::vector<std::string> preqs;
	std::map<int,int> range;

	if (rowRangeVec.size() >  1)
	{
		errstr << "More than one time range while filtering between observation start and observation stop"
				<< Sas::error("MoreThanOneRange");
	}

	if (rowRangeVec.size() ==  0)
	{
		errstr << "Empty time range while filtering between observation start and observation stop."
				<< " Checking earlier stable pointing..." <<
				Sas::warning("EmptyTimeRange");
		STime newStartTime = findClosestValTime(_obsTStart, valtime, valtimeCol->rows(), valdur);
		std::stringstream expr;
		expr << " (VALTIME < \"" << _obsTStop <<  "\") && (VALTIME >= \""<<  newStartTime << "\")";

		DataFilter obsFilter(expr.str());

		errstr << "Filter expression:  " << expr.str() <<  Sas::message(Msg::App, Msg::Sparse);

		obsFilter.doSelect(*atsTab,rowRangeVec);

		if (rowRangeVec.size() ==  0)
			errstr << "Empty time range while filtering between observation start and observation stop"
			<< Sas::error("EmptyTimeRange");
	}

	//Find the first "P" within the vector range
	int firstPindex = 0;
	Column *typeIDc = atsTab->column("TYPEID");
	ColumnData *typeID = typeIDc->data();



	bool poitingFlag = false;
	for (unsigned int i = rowRangeVec[0].startRow ; i <= rowRangeVec[0].stopRow ; ++i)
	{

		string tmp = SasDalUtils::getStringColumnRow(typeID, i);

		if (tmp == "P" )
		{
			firstPindex = i;
			errstr << "First 'P' value found at index: " << i << Sas::message(Msg::App, Msg::Sparse);
			break;
		}

		if (i == rowRangeVec[0].stopRow)
		{
			errstr << "No P TYPEID value found: " << i << Sas::message(Msg::App, Msg::Sparse);
			errstr << "Searching for C or O types... "  << Sas::message(Msg::App, Msg::Sparse);
			poitingFlag = true;
		}
	}

	if (poitingFlag == true)
	{

		for (unsigned int i = rowRangeVec[0].stopRow ; i >= rowRangeVec[0].startRow ; --i)
		{
			string tmp = SasDalUtils::getStringColumnRow(typeID, i);
			if ( (tmp == "C") || (tmp == "O") )
			{
				firstPindex = i+1;
				errstr << "First '" << tmp << "' value found at index: " << i << Sas::message(Msg::App, Msg::Sparse);
				break;
			}
			else
			{
				firstPindex = rowRangeVec[0].startRow;
				poitingFlag = false;
			}
		}
		if (poitingFlag == false)
			errstr << "No P, C or O TYPEID found. We will use the default period "
					<< Sas::message(Msg::App, Msg::Sparse);

	}


	//for (unsigned int i =rowRangeVec[0].startRow ; i <= rowRangeVec[0].stopRow ; ++i)
	for (unsigned int i = firstPindex ; i <= rowRangeVec[0].stopRow ; ++i)
	{

		string tmp = SasDalUtils::getStringColumnRow(cd, i);
		if (preq != tmp )
		{
			preq = tmp;
			preqs.push_back(preq);
		}
	}

	errstr << "Number of PREQs  found: " << preqs.size() <<  Sas::message(Msg::App, Msg::Sparse);


	errstr << "Creating GTI table for each stable pointing..." <<  Sas::message(Msg::App, Msg::Sparse);
	std::map<int,vector<double> > stablePointingMap;



	vector<double> A;
	int pointing = 31;
	for (unsigned int j = 0 ; j < preqs.size() ; ++j)
	{
		stringstream gtiExpr;

		gtiExpr << "(PREQID == \"" << trim(preqs[j]) << "\") && ((TYPEID != \"C\") && (TYPEID != \"O\")) ";

		errstr<< "PREQ filter expression: " << gtiExpr.str() <<Sas::message(Msg::AppLib, Msg::Sparse);

		GtiTableFilter  tableFilter(gtiExpr.str());
		std::stringstream tabName;
		tabName << atsSetName << ":" << extName;

		std::string timeColName = "VALTIME";

		GtiList gtis = tableFilter.gtis(SetTableColumn(tabName.str()),timeColName,
				0.0,
				0.,
				0.0);

		errstr << "PREQ: "<< trim(preqs[j])  << " Number of pointing intervals: "
				<< gtis.size() <<Sas::message(Msg::App, Msg::Sparse);


		// Check if the PREQ belongs to a manual manoeuvre
		if ( (gtis.size() != 0))
		{
			STime start = gtis.front().start() +Xmm::missionReferenceTime;
			STime stop = gtis.back().stop() +Xmm::missionReferenceTime;

			//Get the VALDUR value to substract to the bin.

			unsigned int n = 0;

			//Start time bin
			while ((( std::abs(start.toString(4, STime::FITS) - SasDalUtils::getStringColumnRow(valtime, n) ) > 0.01 ))
					&& (n < valtimeCol->rows()))
			{
				//cout << "TIME "<< n << " " << start.toString(4, STime::FITS) << " " << SasDalUtils::getStringColumnRow(valtime, n)
				//		<< " " <<abs(start.toString(4, STime::FITS) - SasDalUtils::getStringColumnRow(valtime, n) )<< endl;
				n++;
			}

			double dur = valdur[n];

			//Stop time bin
			unsigned int binStop = 0;
			while ((( std::abs(stop.toString(4, STime::FITS) - SasDalUtils::getStringColumnRow(valtime, binStop) ) > 0.01 ))
					&& (binStop < valtimeCol->rows()))
			{
				//cout << "TIME "<< binStop << " " << start.toString(4, STime::FITS) << " " << SasDalUtils::getStringColumnRow(valtime, binStop)
				//		<< " " <<abs(start.toString(4, STime::FITS) - SasDalUtils::getStringColumnRow(valtime, binStop) )<< endl;
				binStop++;
			}
			errstr << "Adding VALDUR = " << valdur[binStop] << " to the stable pointing stop time"<<Sas::message(Msg::App, Msg::Sparse);
			double stopDur = valdur[binStop];
			stop = stop + stopDur;

			if ( (start == stop) || (stop < (start+dur) ) )
			{
				errstr << "identical Start and Stop values... taking into account VALDUR value." <<Sas::message(Msg::App, Msg::Sparse);
				stop = stop+dur;
			}
			double stopVal = stop - Xmm::missionReferenceTime;

			//Check if the stable pointing start time happends before the observation start. In this case, we set the stable pointin
			// start time to the observation start

			if (start < _obsTStart)
			{
				errstr << "Setting start time to the obs start... "<< _obsTStart.toString()<<Sas::message(Msg::App, Msg::Sparse);
				start = _obsTStart;
			}
			double startVal = start -  Xmm::missionReferenceTime;

			//Check the pointing corrdinates
			AttitudeLocator *att = new  AttitudeLocator();
			att->setAHF(atsSetName,start,stop);
			vector<double> med = att->getMedianAtt();
			delete att;
			//cout <<"MEDIAN RA " << med[0]*RAD_TO_DEG<< " DEC: " << med[1]*RAD_TO_DEG << endl;

			if ( (preqs[j].substr(9,1) == "9") && (stablePointingMap.size() > 0)  )
			{
				errstr << "Manual manoeuvre found --> PREQ: " << preqs[j]  <<Sas::message(Msg::App, Msg::Sparse);

				//Check angular separation
				double angDist =  angularSeparation(A,med)*RAD_TO_DEG;
				errstr << "ANGULAR DISTANCE: " << angDist <<Sas::message(Msg::App, Msg::Sparse);
				errstr << "Manual manoeuvre Time interval " <<gtis.back().stop()-gtis.back().start() <<Sas::message(Msg::App, Msg::Sparse);
				A = med;
				if (angDist < 0.2)
				{
					errstr << "Pointing added to previous pointing"  <<Sas::message(Msg::App, Msg::Sparse);
					std::vector<double> val;
					std::map<int,vector<double> >::iterator it =stablePointingMap.end();
					--it;
					val.push_back((*it).second.front());
					val.push_back(stopVal);
					stablePointingMap[(*it).first] = val;

				}
				else
				{
					errstr << "New Pointing added. PREQ " <<  preqs[j] <<Sas::message(Msg::App, Msg::Sparse);
					std::vector<double> val;
					val.push_back(startVal);
					val.push_back(stopVal);
					stablePointingMap[pointing] = val;
					pointing++;
				}
			}
			else
			{
				//<< " START: " << gtis.front().start()+Xmm::missionReferenceTime
				//<< " STOP: " << gtis.back().stop()+Xmm::missionReferenceTime <<Sas::message(Msg::App, Msg::Sparse);
				errstr << "New pointing added. PREQ: " <<  preqs[j] <<Sas::message(Msg::App, Msg::Sparse);
				std::vector<double> val;
				val.push_back(startVal);
				val.push_back(stopVal);
				stablePointingMap[pointing] = val;
				pointing++;
				A = med;

			}
		}
	}
	dataSetServer->close(s);
	//c->release();
	//delete cd;
	//delete valtime;
	dataSetServer->close(atsDS);

	dataSetServer->remove(atsSetName);



	DataSet *set = dataSetServer->open(stringParameter("preqgtifile"),Dal::Create);

	int size = stablePointingMap.size();

	Table *tab = set->addTable("STDGTI",size);

	ColumnData *startCol = tab->addColumn("START", Dal::Real64,"","Time(s)") ->data();
	ColumnData *stopCol = tab->addColumn("STOP",Dal::Real64,"","Time(s)")->data();
	real64 *startData=0;
	real64 *stopData=0;
	startData = startCol->real64Data();
	stopData = stopCol->real64Data();



	unsigned int count = 0;
	for (std::map<int,vector<double> >::iterator i = stablePointingMap.begin() ;
			i != stablePointingMap.end() ; ++i)
	{
		errstr << "PREQ " <<(*i).first <<": START "<< ((*i).second)[0]+Xmm::missionReferenceTime
				<< " STOP " << ((*i).second)[1]+Xmm::missionReferenceTime << Sas::message(Msg::App, Msg::Sparse);

		startData[count] =  ((*i).second)[0];
		stopData[count] = ((*i).second)[1];
		count++;

	}

	delete startCol;
	delete stopCol;

	dataSetServer->close(set);


}

STime findClosestValTime(STime obsTStart, ColumnData* valTime,unsigned rows, real32* valDur)
{
	unsigned int count=0;
	for (unsigned int i = 0 ; i <= rows-1; ++i)
	{
		STime t(SasDalUtils::getStringColumnRow(valTime, i) );
		if ( (obsTStart - t) <= 0.0)
		{
			count = i-1;
			break;
		}
		if (i == rows-1)
		{
			errstr << "None VALTIME value found close the exposure start time (" << obsTStart.toString() << ")"
					<< Sas::error("VALTIMENotFound");
		}

	}

	STime tmpTime(SasDalUtils::getStringColumnRow(valTime, count));



	if (tmpTime+valDur[count] > _obsTStop)
		count--;
	STime newStartTime(SasDalUtils::getStringColumnRow(valTime, count));

	if ( (tmpTime < obsTStart) && (tmpTime+valDur[count] > obsTStart) )
		errstr << "Exposure Start Time: " << obsTStart.toString() << "is contained in the row " << count<< " with the following VALTIME: "
			<< newStartTime.toString() << " and VALDUR "<< valDur[count] << endl;

	errstr << "Exposure Start Time: " << obsTStart.toString() << " Closest VALTIME: " << newStartTime.toString() << endl;
	return newStartTime;

}



void addAttitude()
{


	Info::Task* atthkgen = task("atthkgen");
	ParameterInfo& atthkgenpars = *atthkgen->parameters();
	atthkgenpars.revertToDefault();
	atthkgenpars["withpreqgti"]->setValue(true);
	atthkgenpars["preqgtifile"]->setValue(stringParameter("preqgtifile"));
	string attoutName = tmpnam(0);
	atthkgenpars["atthkset"]->setValue(attoutName);

	atthkgen->exec();
	unlink(attoutName.c_str());
}

	string _stableGTITableName;
private:
	string _ATSFile;
	STime _obsTStart;
	STime _obsTStop;


};





void preqgti ()
{

	//Stable pointing GTI filtering
	//SOC-SCR-336: Attfilter to filter attitude according to pointing-request type.
	//This new functionality creates a GTI file containing the stable pointing
	//periods in the ATS file.
	//	instantiate stable pointing object ...

	StablePointngGtiTable *stPointing;
	if ( booleanParameter("withatsfile") && !booleanParameter("withtimeranges") )
		stPointing = new StablePointngGtiTable(stringParameter("atsfile"));
	else if ( booleanParameter("withatsfile") && booleanParameter("withtimeranges") )
		stPointing = new StablePointngGtiTable( stringParameter("atsfile"),realParameter("timebegin"),realParameter("timeend") );
	else if ( !booleanParameter("withatsfile") && booleanParameter("withtimeranges") )
		stPointing = new StablePointngGtiTable(realParameter("timebegin"),realParameter("timeend") );
	else
		stPointing = new StablePointngGtiTable();

	stPointing->getPReqValues();

	if (booleanParameter("withpointingattitude"))
		stPointing->addAttitude();

	//stPointing.createGtiTable();


	delete stPointing;

} //preqgti
